@@ -21,7 +21,6 @@
 
 #include <config.h>
 #include <gtk/gtk.h>
-#include <glade/glade.h>
 #include <string.h>
 #include <stdlib.h>
 #include "scrollarea.h"
@@ -119,22 +118,12 @@ idle_free (gchar *s)
     return s;
 }
 
-static int
-compare_outputs (const void *p1, const void *p2)
-{
-    GnomeOutputInfo *const *o1 = p1;
-    GnomeOutputInfo *const *o2 = p2;
-
-    return (**o1).x - (**o2).x;
-}
-
 static void
 on_screen_changed (GnomeRRScreen *scr,
 		   gpointer data)
 {
     GnomeRRConfig *current;
     App *app = data;
-    int i;
 
     current = gnome_rr_config_new_current (app->screen);
 
@@ -144,25 +133,6 @@ on_screen_changed (GnomeRRScreen *scr,
     app->current_configuration = current;
     app->current_output = NULL;
 
-#if 0
-    for (i = 0; app->current_configuration->outputs[i] != NULL; ++i)
-    {
-	GnomeOutputInfo *o = app->current_configuration->outputs[i];
-
-	g_debug ("  output %s %s: %d %d %d %d", o->name, o->on? "on" : "off", o->x, o->y, o->width, o->height);
-    }
-#endif
-
-#if 0
-    g_debug ("sorting");
-#endif
-    /* Sort outputs according to X coordinate */
-    for (i = 0; app->current_configuration->outputs[i] != NULL; ++i)
-	;
-
-    qsort (app->current_configuration->outputs, i, sizeof (GnomeOutputInfo *),
-	   compare_outputs);
-
     if (app->labeler) {
 	gnome_rr_labeler_hide (app->labeler);
 	g_object_unref (app->labeler);
@@ -170,15 +140,6 @@ on_screen_changed (GnomeRRScreen *scr,
 
     app->labeler = gnome_rr_labeler_new (app->current_configuration);
 
-#if 0
-    for (i = 0; app->current_configuration->outputs[i] != NULL; ++i)
-    {
-	GnomeOutputInfo *o = app->current_configuration->outputs[i];
-
-	g_debug ("  output: %d %d %d %d", o->x, o->y, o->width, o->height);
-    }
-#endif
-
     select_current_output_from_dialog_position (app);
 }
 
@@ -379,6 +340,12 @@ rebuild_rotation_combo (App *app)
 	combo_select (app->rotation_combo, _("Normal"));
 }
 
+static char *
+make_rate_string (int hz)
+{
+    return g_strdup_printf (_("%d Hz"), hz);
+}
+
 static void
 rebuild_rate_combo (App *app)
 {
@@ -413,7 +380,7 @@ rebuild_rate_combo (App *app)
 	    height == app->current_output->height)
 	{
 	    add_key (app->refresh_combo,
-		     idle_free (g_strdup_printf (_("%d Hz"), rate)),
+		     idle_free (make_rate_string (rate)),
 		     0, 0, rate, -1);
 
 	    if (rate > best)
@@ -421,8 +388,8 @@ rebuild_rate_combo (App *app)
 	}
     }
 
-    if (!combo_select (app->refresh_combo, idle_free (g_strdup_printf (_("%d Hz"), app->current_output->rate))))
-	combo_select (app->refresh_combo, idle_free (g_strdup_printf (_("%d Hz"), best)));
+    if (!combo_select (app->refresh_combo, idle_free (make_rate_string (app->current_output->rate))))
+	combo_select (app->refresh_combo, idle_free (make_rate_string (best)));
 }
 
 static int
@@ -456,29 +423,26 @@ count_all_outputs (GnomeRRConfig *config)
 static void
 rebuild_current_monitor_label (App *app)
 {
-	char *str;
-	gboolean free_str;
+	char *str, *tmp;
 	GdkColor color;
 	gboolean use_color;
 
 	if (app->current_output)
 	{
-	    str = g_strdup_printf (_("<b>Monitor: %s</b>"), app->current_output->display_name);
-	    free_str = TRUE;
+	    tmp = g_strdup_printf (_("Monitor: %s"), app->current_output->display_name);
+	    str = g_strdup_printf ("<b>%s</b>", tmp);
 	    gnome_rr_labeler_get_color_for_output (app->labeler, app->current_output, &color);
 	    use_color = TRUE;
+	    g_free (tmp);
 	}
 	else
 	{
-	    str = _("<b>Monitor</b>");
-	    free_str = FALSE;
+	    str = g_strdup_printf ("<b>%s</b>", _("Monitor"));
 	    use_color = FALSE;
 	}
 
 	gtk_label_set_markup (GTK_LABEL (app->current_monitor_label), str);
-
-	if (free_str)
-	    g_free (str);
+	g_free (str);
 
 	if (use_color)
 	{
@@ -541,12 +505,17 @@ rebuild_on_off_radios (App *app)
     g_signal_handlers_unblock_by_func (app->monitor_off_radio, G_CALLBACK (monitor_on_off_toggled_cb), app);
 }
 
+static char *
+make_resolution_string (int width, int height)
+{
+    return g_strdup_printf (_("%d x %d"), width, height);
+}
+
 static void
 rebuild_resolution_combo (App *app)
 {
     int i;
     GnomeRRMode **modes;
-    int best_w, best_h;
     const char *current;
 
     clear_combo (app->resolution_combo);
@@ -560,11 +529,10 @@ rebuild_resolution_combo (App *app)
     }
 
     g_assert (app->current_output != NULL);
+    g_assert (app->current_output->width != 0 && app->current_output->height != 0);
 
     gtk_widget_set_sensitive (app->resolution_combo, TRUE);
 
-    best_w = 0;
-    best_h = 0;
     for (i = 0; modes[i] != NULL; ++i)
     {
 	int width, height;
@@ -573,26 +541,14 @@ rebuild_resolution_combo (App *app)
 	height = gnome_rr_mode_get_height (modes[i]);
 
 	add_key (app->resolution_combo,
-		 idle_free (g_strdup_printf (_("%d x %d"), width, height)),
+		 idle_free (make_resolution_string (width, height)),
 		 width, height, 0, -1);
-
-	if (width * height > best_w * best_h)
-	{
-	    best_w = width;
-	    best_h = height;
-	}
     }
 
-    current = idle_free (g_strdup_printf (_("%d x %d"),
-					  app->current_output->width,
-					  app->current_output->height));
+    current = idle_free (make_resolution_string (app->current_output->width, app->current_output->height));
 
     if (!combo_select (app->resolution_combo, current))
-    {
-	combo_select (app->resolution_combo,
-		      idle_free (
-			  g_strdup_printf (_("%d x %d"), best_w, best_h)));
-    }
+	g_assert_not_reached ();
 }
 
 static void
@@ -697,6 +653,52 @@ on_rate_changed (GtkComboBox *box, gpointer data)
     foo_scroll_area_invalidate (FOO_SCROLL_AREA (app->area));
 }
 
+static void
+find_best_mode (GnomeRRMode **modes, int *out_width, int *out_height)
+{
+    int i;
+
+    *out_width = 0;
+    *out_height = 0;
+
+    for (i = 0; modes[i] != NULL; i++)
+    {
+	int w, h;
+
+	w = gnome_rr_mode_get_width (modes[i]);
+	h = gnome_rr_mode_get_height (modes[i]);
+
+	if (w * h > *out_width * *out_height)
+	{
+	    *out_width = w;
+	    *out_height = h;
+	}
+    }
+}
+
+static void
+select_resolution_for_current_output (App *app)
+{
+    GnomeRRMode **modes;
+    int width, height;
+
+    if (app->current_output->pref_width != 0 && app->current_output->pref_height != 0)
+    {
+	app->current_output->width = app->current_output->pref_width;
+	app->current_output->height = app->current_output->pref_height;
+	return;
+    }
+
+    modes = get_current_modes (app);
+    if (!modes)
+	return;
+
+    find_best_mode (modes, &width, &height);
+
+    app->current_output->width = width;
+    app->current_output->height = height;
+}
+
 static void
 monitor_on_off_toggled_cb (GtkToggleButton *toggle, gpointer data)
 {
@@ -721,6 +723,9 @@ monitor_on_off_toggled_cb (GtkToggleButton *toggle, gpointer data)
 
     app->current_output->on = is_on;
 
+    if (is_on)
+	select_resolution_for_current_output (app); /* The refresh rate will be picked in rebuild_rate_combo() */
+
     rebuild_gui (app);
     foo_scroll_area_invalidate (FOO_SCROLL_AREA (app->area));
 }
@@ -740,6 +745,9 @@ realign_outputs_after_resolution_change (App *app, GnomeOutputInfo *output_that_
 
     g_assert (app->current_configuration != NULL);
 
+    if (output_that_changed->width == old_width && output_that_changed->height == old_height)
+	return;
+
     old_right_edge = output_that_changed->x + old_width;
     old_bottom_edge = output_that_changed->y + old_height;
 
@@ -1492,11 +1500,13 @@ paint_output (App *app, cairo_t *cr, int i)
     GList *connected_outputs = list_connected_outputs (app, &total_w, &total_h);
     GnomeOutputInfo *output = g_list_nth (connected_outputs, i)->data;
     PangoLayout *layout = get_display_name (app, output);
-    PangoRectangle extent;
+    PangoRectangle ink_extent, log_extent;
     GdkRectangle viewport;
     double angle;
     GdkColor output_color;
     double r, g, b;
+    double available_w;
+    double factor;
 
     cairo_save (cr);
 
@@ -1603,13 +1613,19 @@ paint_output (App *app, cairo_t *cr, int i)
     cairo_set_line_width (cr, 2);
 
     layout_set_font (layout, "Sans Bold 12");
+    pango_layout_get_pixel_extents (layout, &ink_extent, &log_extent);
 
-    pango_layout_get_pixel_extents (layout, NULL, &extent);
+    available_w = w * scale + 0.5 - 6; /* Same as the inner rectangle's width, minus 1 pixel of padding on each side */
+    if (available_w < ink_extent.width)
+	factor = available_w / ink_extent.width;
+    else
+	factor = 1.0;
 
-    extent.x = x + ((w * scale + 0.5) - extent.width) / 2;
-    extent.y = y + ((h * scale + 0.5) - extent.height) / 2;
+    cairo_move_to (cr,
+		   x + ((w * scale + 0.5) - factor * log_extent.width) / 2,
+		   y + ((h * scale + 0.5) - factor * log_extent.height) / 2);
 
-    cairo_move_to (cr, extent.x, extent.y);
+    cairo_scale (cr, factor, factor);
 
     if (output->on)
 	cairo_set_source_rgb (cr, 0.0, 0.0, 0.0);
@@ -2110,38 +2126,47 @@ apply_button_clicked_cb (GtkButton *button, gpointer data)
     app->apply_button_clicked_timestamp = gtk_get_current_event_time ();
 }
 
+static GtkWidget*
+_gtk_builder_get_widget (GtkBuilder *builder, const gchar *name)
+{
+    return GTK_WIDGET (gtk_builder_get_object (builder, name));
+}
+
 static void
 run_application (App *app)
 {
-#ifndef GLADEDIR
-#define GLADEDIR "."
+#ifndef UIDIR
+#define UIDIR "."
 #endif
-#define GLADE_FILE GLADEDIR "/display-capplet.glade"
-    GladeXML *xml;
+#define UI_FILE UIDIR "/display-capplet.ui"
+    GtkBuilder *builder;
     GtkWidget *align;
     GError *error;
 
-    xml = glade_xml_new (GLADE_FILE, NULL, NULL);
-    if (!xml)
+    error = NULL;
+    builder = gtk_builder_new ();
+
+    if (gtk_builder_add_from_file (builder, UI_FILE, &error) == 0)
     {
-	g_warning ("Could not open " GLADE_FILE);
+	g_warning ("Could not parse UI definition: %s", error->message);
+	g_error_free (error);
+	g_object_unref (builder);
 	return;
     }
 
-    error = NULL;
     app->screen = gnome_rr_screen_new (gdk_screen_get_default (),
 				       on_screen_changed, app, &error);
     if (!app->screen)
     {
 	error_message (NULL, _("Could not get screen information"), error->message);
 	g_error_free (error);
-	g_object_unref (xml);
+	g_object_unref (builder);
 	return;
     }
 
     app->client = gconf_client_get_default ();
 
-    app->dialog = glade_xml_get_widget (xml, "dialog");
+    app->dialog = _gtk_builder_get_widget (builder, "dialog");
     g_signal_connect_after (app->dialog, "map-event",
 			    G_CALLBACK (dialog_map_event_cb), app);
 
@@ -2149,43 +2174,49 @@ run_application (App *app)
     gtk_window_set_icon_name (GTK_WINDOW (app->dialog),
 			      "gnome-display-properties");
 
-    app->current_monitor_event_box = glade_xml_get_widget (xml, "current_monitor_event_box");
-    app->current_monitor_label = glade_xml_get_widget (xml, "current_monitor_label");
+    app->current_monitor_event_box = _gtk_builder_get_widget (builder,
+    						   "current_monitor_event_box");
+    app->current_monitor_label = _gtk_builder_get_widget (builder,
+    						       "current_monitor_label");
 
-    app->monitor_on_radio = glade_xml_get_widget (xml, "monitor_on_radio");
-    app->monitor_off_radio = glade_xml_get_widget (xml, "monitor_off_radio");
+    app->monitor_on_radio = _gtk_builder_get_widget (builder,
+    						     "monitor_on_radio");
+    app->monitor_off_radio = _gtk_builder_get_widget (builder,
+    						      "monitor_off_radio");
     g_signal_connect (app->monitor_on_radio, "toggled",
 		      G_CALLBACK (monitor_on_off_toggled_cb), app);
     g_signal_connect (app->monitor_off_radio, "toggled",
 		      G_CALLBACK (monitor_on_off_toggled_cb), app);
 
-    app->resolution_combo = glade_xml_get_widget (xml, "resolution_combo");
+    app->resolution_combo = _gtk_builder_get_widget (builder,
+    						     "resolution_combo");
     g_signal_connect (app->resolution_combo, "changed",
 		      G_CALLBACK (on_resolution_changed), app);
 
-    app->refresh_combo = glade_xml_get_widget (xml, "refresh_combo");
+    app->refresh_combo = _gtk_builder_get_widget (builder, "refresh_combo");
     g_signal_connect (app->refresh_combo, "changed",
 		      G_CALLBACK (on_rate_changed), app);
 
-    app->rotation_combo = glade_xml_get_widget (xml, "rotation_combo");
+    app->rotation_combo = _gtk_builder_get_widget (builder, "rotation_combo");
     g_signal_connect (app->rotation_combo, "changed",
 		      G_CALLBACK (on_rotation_changed), app);
 
-    app->clone_checkbox = glade_xml_get_widget (xml, "clone_checkbox");
+    app->clone_checkbox = _gtk_builder_get_widget (builder, "clone_checkbox");
     g_signal_connect (app->clone_checkbox, "toggled",
 		      G_CALLBACK (on_clone_changed), app);
 
-    g_signal_connect (glade_xml_get_widget (xml, "detect_displays_button"),
+    g_signal_connect (_gtk_builder_get_widget (builder, "detect_displays_button"),
 		      "clicked", G_CALLBACK (on_detect_displays), app);
 
-    app->show_icon_checkbox = glade_xml_get_widget (xml, "show_notification_icon");
+    app->show_icon_checkbox = _gtk_builder_get_widget (builder,
+						      "show_notification_icon");
 
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (app->show_icon_checkbox),
 				  gconf_client_get_bool (app->client, SHOW_ICON_KEY, NULL));
 
     g_signal_connect (app->show_icon_checkbox, "toggled", G_CALLBACK (on_show_icon_toggled), app);
 
-    app->panel_checkbox = glade_xml_get_widget (xml, "panel_checkbox");
+    app->panel_checkbox = _gtk_builder_get_widget (builder, "panel_checkbox");
 
     make_text_combo (app->resolution_combo, 4);
     make_text_combo (app->refresh_combo, 3);
@@ -2206,20 +2237,20 @@ run_application (App *app)
     g_signal_connect (app->area, "viewport_changed",
 		      G_CALLBACK (on_viewport_changed), app);
 
-    align = glade_xml_get_widget (xml, "align");
+    align = _gtk_builder_get_widget (builder, "align");
 
     gtk_container_add (GTK_CONTAINER (align), app->area);
 
     /* Until we have help to show, we'll just hide the Help button */
     hide_help_button (app);
 
-    app->apply_button = glade_xml_get_widget (xml, "apply_button");
+    app->apply_button = _gtk_builder_get_widget (builder, "apply_button");
     g_signal_connect (app->apply_button, "clicked",
 		      G_CALLBACK (apply_button_clicked_cb), app);
 
     on_screen_changed (app->screen, app);
 
-    g_object_unref (xml);
+    g_object_unref (builder);
 
 restart:
     switch (gtk_dialog_run (GTK_DIALOG (app->dialog)))
